@@ -35,12 +35,14 @@
 
 /* Xserver interface */
 #include "xf86.h"
+#if HAVE_XF86_ANSIC_H
+# include "xf86_ansic.h" 
+#endif
 
 /* Driver specific headers */
 #include "rhd.h"
 #include "rhd_randr.h"
 
-
 #if (RANDR_MAJOR == 1 && RANDR_MINOR >= 2) || RANDR_MAJOR >= 2
 
 
@@ -56,31 +58,33 @@
 /* Driver specific headers */
 #include "rhd.h"
 #include "rhd_crtc.h"
+#include "rhd_cursor.h"
 #include "rhd_connector.h"
 #include "rhd_output.h"
 #include "rhd_modes.h"
 #include "rhd_monitor.h"
 #include "rhd_vga.h"
 #include "rhd_pll.h"
+#include "rhd_lut.h"
 #include "rhd_mc.h"
 #include "rhd_card.h"
 
-/* System headers */
-#ifndef _XF86_ANSIC_H
-#include <ctype.h>
-#include <string.h>
-#include <stdio.h>
-#endif
-
-
 /*
  * Driver internal
  */
 
-/* List of allocated RandR Crtc and Output description structs */
+#define WRAP_SCRNINFO(func,new)					\
+    do {							\
+	rhdPtr->randr->func = pScrn->func;			\
+	pScrn->func = new;					\
+    } while (0)
+#define UNWRAP_SCRNINFO(func)					\
+	pScrn->func = rhdPtr->randr->func
+
 struct rhdRandr {
     xf86CrtcPtr    RandrCrtc[2];
     xf86OutputPtr *RandrOutput;  /* NULL-terminated */
+    void         (*PointerMoved)(int, int, int);
 } ;
 
 /* Outputs and Connectors are combined for RandR due to missing abstraction */
@@ -94,9 +98,10 @@ typedef struct _rhdRandrOutput {
 #define ATOM_CONNECTOR_TYPE   "RANDR_CONNECTOR_TYPE"
 #define ATOM_CONNECTOR_NUMBER "RANDR_CONNECTOR_NUMBER"
 #define ATOM_OUTPUT_NUMBER    "RANDR_OUTPUT_NUMBER"
+#define ATOM_PANNING_AREA     "RANDR_PANING_AREA"
 
 static Atom atomSignalFormat, atomConnectorType, atomConnectorNumber,
-	    atomOutputNumber;
+	    atomOutputNumber, atomPanningArea;
 
 
 /* Get RandR property values */
@@ -153,6 +158,27 @@ rhdGetConnectorType(rhdRandrOutputPtr ro)
     }
 }
 
+/* Set crtc pos according to mouse pos and panning information */
+static void
+rhdUpdateCrtcPos(struct rhdCrtc *Crtc, int x, int y)
+{
+    if (Crtc->MaxX > 0) {
+	int cx = Crtc->X, cy = Crtc->Y;
+	int w  = Crtc->CurrentMode->HDisplay;
+	int h  = Crtc->CurrentMode->VDisplay;
+	if (x < cx)
+	    cx = x > Crtc->MinX ? x : Crtc->MinX;
+	if (x >= cx + w)
+	    cx = x < Crtc->MaxX ? x-w+1 : Crtc->MaxX-w;
+	if (y < cy)
+	    cy = y > Crtc->MinY ? y : Crtc->MinY;
+	if (y >= cy + h)
+	    cy = y < Crtc->MaxY ? y-h+1 : Crtc->MaxY-h;
+	if (cx != Crtc->X || cy != Crtc->Y)
+	    Crtc->FrameSet(Crtc, cx, cy);
+    }
+}
+
 /* Debug: print out state */
 void
 RHDDebugRandrState (RHDPtr rhdPtr, const char *msg)
@@ -295,6 +321,12 @@ rhdRRCrtcPrepare(xf86CrtcPtr crtc)
 
     /* Disable CRTCs to stop noise from appearing. */
     Crtc->Power(Crtc, RHD_POWER_RESET);
+
+    /* Verify panning area */
+    if (Crtc->MaxX > Crtc->Width)
+	Crtc->MaxX = Crtc->Width;
+    if (Crtc->MaxY > Crtc->Height)
+	Crtc->MaxY = Crtc->Height;
 }
 static void
 rhdRRCrtcModeSet(xf86CrtcPtr  crtc,
@@ -317,6 +349,7 @@ rhdRRCrtcModeSet(xf86CrtcPtr  crtc,
 		pScrn->depth, rhdPtr->FbFreeStart);
     Crtc->ModeSet(Crtc, Mode);
     Crtc->FrameSet(Crtc, x, y);
+    rhdUpdateCrtcPos(Crtc, Crtc->Cursor->X, Crtc->Cursor->Y);
     RHDPLLSet(Crtc->PLL, Mode->Clock);		/* This also powers up PLL */
     Crtc->PLLSelect(Crtc, Crtc->PLL);
     Crtc->LUTSelect(Crtc, Crtc->LUT);
@@ -335,6 +368,36 @@ rhdRRCrtcCommit(xf86CrtcPtr crtc)
     RHDDebugRandrState(rhdPtr, Crtc->Name);
 }
 
+/*
+ * They just had to do NIH again here: Old X functionality provides a size, a
+ * list of indices, and a table of RGB unsigned shorts. RandR provides what
+ * is below. Apart from horribly breaking any attempt at being backwards
+ * compatible, this also pretty much rules out the usage of indexed colours, as
+ * each time even a single colour is changed an entirely new table has to be
+ * uploaded. Just cute. -- libv.
+ */
+static void
+rhdRRCrtcGammaSet(xf86CrtcPtr crtc, CARD16 *red, CARD16 *green, CARD16 *blue,
+		  int size)
+{
+    struct rhdCrtc *Crtc = (struct rhdCrtc *) crtc->driver_private;
+    int indices[0x100]; /* would RandR use a size larger than 256? */
+    LOCO colors[0x100];
+    int i;
+
+    RHDDebug(Crtc->scrnIndex, "%s: %s.\n", __func__, Crtc->Name);
+
+    /* thanks so very much */
+    for (i = 0; i < size; i++) {
+	indices[i] = i;
+	colors[i].red = red[i];
+	colors[i].green = green[i];
+	colors[i].blue = blue[i];
+    }
+
+    Crtc->LUT->Set(Crtc->LUT, size, indices, colors);
+}
+
 /* Dummy, because not tested for NULL */
 static Bool
 rhdRRCrtcModeFixupDUMMY(xf86CrtcPtr    crtc, 
@@ -383,13 +446,28 @@ rhdRROutputCreateResources(xf86OutputPtr out)
     CARD32            num;
 
     RHDFUNC(rhdPtr);
-    /* Set up potential RandR 1.3 properties */
+
+    /* Create atoms for potential RandR 1.3 properties */
+    atomSignalFormat    = MakeAtom(ATOM_SIGNAL_FORMAT,
+				   sizeof(ATOM_SIGNAL_FORMAT)-1, TRUE);
+    atomConnectorType   = MakeAtom(ATOM_CONNECTOR_TYPE,
+				   sizeof(ATOM_CONNECTOR_TYPE)-1, TRUE);
+    atomConnectorNumber = MakeAtom(ATOM_CONNECTOR_NUMBER,
+				   sizeof(ATOM_CONNECTOR_NUMBER)-1, TRUE);
+    atomOutputNumber    = MakeAtom(ATOM_OUTPUT_NUMBER,
+				   sizeof(ATOM_OUTPUT_NUMBER)-1, TRUE);
+    atomPanningArea     = MakeAtom(ATOM_PANNING_AREA,
+				   sizeof(ATOM_PANNING_AREA)-1, TRUE);
+
+    /* Set up properties */
     RRConfigureOutputProperty(out->randr_output, atomSignalFormat,
 			      FALSE, FALSE, TRUE, 0, NULL);
     RRConfigureOutputProperty(out->randr_output, atomConnectorType,
 			      FALSE, FALSE, TRUE, 0, NULL);
     RRConfigureOutputProperty(out->randr_output, atomConnectorNumber,
 			      FALSE, FALSE, TRUE, 0, NULL);
+    RRConfigureOutputProperty(out->randr_output, atomPanningArea,
+			      FALSE, FALSE, FALSE, 0, NULL);
     val = rhdGetSignalFormat(rout);
     RRChangeOutputProperty(out->randr_output, atomSignalFormat,
 			   XA_STRING, 8, PropModeReplace,
@@ -413,6 +491,9 @@ rhdRROutputCreateResources(xf86OutputPtr out)
     RRChangeOutputProperty(out->randr_output, atomOutputNumber,
 			   XA_INTEGER, 32, PropModeReplace,
 			   1, &num, FALSE, FALSE);
+    RRChangeOutputProperty(out->randr_output, atomPanningArea,
+			   XA_STRING, 8, PropModeReplace,
+			   0, NULL, FALSE, FALSE);
 }
 
 /* Turns the output on/off, or sets intermediate power levels if available. */
@@ -625,8 +706,11 @@ rhdRROutputCommit(xf86OutputPtr out)
     RHDPtr            rhdPtr = RHDPTR(out->scrn);
     rhdRandrOutputPtr rout   = (rhdRandrOutputPtr) out->driver_private;
     const char       *val;
+    char              buf[32];
+    struct rhdCrtc   *Crtc   = (struct rhdCrtc *) out->crtc->driver_private;
 
     RHDFUNC(rhdPtr);
+    ASSERT(Crtc == rout->Output->Crtc);
 
     rout->Output->Active    = TRUE;
     rout->Output->Connector = rout->Connector;
@@ -637,6 +721,16 @@ rhdRROutputCommit(xf86OutputPtr out)
     RRChangeOutputProperty(out->randr_output, atomConnectorType,
 			   XA_STRING, 8, PropModeReplace,
 			   strlen(val), (char *) val, TRUE, FALSE);
+    /* Should be a crtc property */
+    if (Crtc->MaxX > Crtc->MinX && Crtc->MaxY > Crtc->MinY)
+	sprintf(buf, "%dx%d+%d+%d", Crtc->MaxX - Crtc->MinX,
+		Crtc->MaxY - Crtc->MinY, Crtc->MinX, Crtc->MinY);
+    else
+	buf[0] = 0;
+    RRChangeOutputProperty(out->randr_output, atomPanningArea,
+			   XA_STRING, 8, PropModeReplace,
+			   strlen(buf), buf, TRUE, FALSE);
+
     RHDDebugRandrState(rhdPtr, rout->Name);
 }
 
@@ -689,16 +783,9 @@ rhdRROutputDetect(xf86OutputPtr output)
 	    /*
 	     * HPD returned false
 	     */
-	    if (rhdPtr->ignoreHpd.set) {
-		if (rout->Output->Sense &&
-		    rout->Output->Sense(rout->Output, rout->Connector->Type))
-		    return XF86OutputStatusConnected;
-	    }
 	    /* There is the infamous DMS-59 connector, on which HPD returns
 	     * false when 'only' VGA is connected. */
-	    else if (rhdPtr->Card &&
-		     (rhdPtr->Card->flags & RHD_CARD_FLAG_DMS59) &&
-		     rout->Connector->Type == RHD_CONNECTOR_VGA) {
+	    if (rhdPtr->Card && (rhdPtr->Card->flags & RHD_CARD_FLAG_DMS59)) {
 		xf86DrvMsg(rhdPtr->scrnIndex, X_INFO,
 			   "RandR: Verifying state of DMS-59 VGA connector.\n");
 		if (rout->Output->Sense &&
@@ -717,6 +804,13 @@ rhdRROutputDetect(xf86OutputPtr output)
 	    else
 		return XF86OutputStatusDisconnected;
 	}
+	/* Use DDC address probing if possible otherwise */
+	if (rout->Connector->DDC) {
+	    if (xf86I2CProbeAddress(rout->Connector->DDC, 0xa0))
+		return XF86OutputStatusConnected;
+	    else
+		return XF86OutputStatusDisconnected;
+	}
 	return XF86OutputStatusUnknown;
     }
 }
@@ -768,11 +862,40 @@ rhdRROutputGetModes(xf86OutputPtr output)
 
 /* An output's property has changed. */
 static Bool
-rhdRROutputSetProperty(xf86OutputPtr output,
-		       Atom property,
+rhdRROutputSetProperty(xf86OutputPtr out, Atom property,
 		       RRPropertyValuePtr value)
 {
-    return FALSE;	/* We don't have mutable properties yet */
+    RHDPtr            rhdPtr = RHDPTR(out->scrn);
+    rhdRandrOutputPtr rout   = (rhdRandrOutputPtr) out->driver_private;
+
+    if (property == atomPanningArea) {
+	int w = 0, h = 0, x = 0, y = 0;
+	struct rhdCrtc *Crtc = rout->Output->Crtc;
+	if (!Crtc)
+	    return FALSE;
+	if (value->type != XA_STRING || value->format != 8)
+	    return FALSE;
+	switch (sscanf(value->data, "%dx%d+%d+%d", &w, &h, &x, &y)) {
+	case 0:
+	case 2:
+	case 4:
+	    if (w < 0 || h < 0 || x < 0 || y < 0 ||
+		x+w > Crtc->Width || y+h > Crtc->Height)
+		return FALSE;
+	    Crtc->MinX = x;
+	    Crtc->MinY = y;
+	    Crtc->MaxX = x + w;
+	    Crtc->MaxY = y + h;
+	    rhdUpdateCrtcPos(Crtc, Crtc->Cursor->X, Crtc->Cursor->Y);
+	    RHDDebug(rhdPtr->scrnIndex, "%s: PanningArea %d/%d - %d/%d\n",
+		     x, y, x+w, y+h);
+	    return TRUE;
+	default:
+	    return FALSE;
+	}
+    }
+
+    return FALSE;	/* Others are not mutable */
 }
 
 
@@ -790,7 +913,7 @@ static const xf86CrtcFuncsRec rhdRRCrtcFuncs = {
     rhdRRCrtcLock, rhdRRCrtcUnlock,
     rhdRRCrtcModeFixupDUMMY,
     rhdRRCrtcPrepare, rhdRRCrtcModeSet, rhdRRCrtcCommit,
-    NULL,						/* CrtcGammaSet */
+    rhdRRCrtcGammaSet,
     /* rhdRRCrtcShadowAllocate,rhdRRCrtcShadowCreate,rhdRRCrtcShadowDestroy */
     NULL, NULL, NULL,
     /* SetCursorColors,SetCursorPosition,ShowCursor,HideCursor,
@@ -919,16 +1042,6 @@ RHDRandrPreInit(ScrnInfoPtr pScrn)
 	return FALSE;
     }
 
-    /* Create atoms for potential RandR 1.3 properties */
-    atomSignalFormat    = MakeAtom(ATOM_SIGNAL_FORMAT,
-				   sizeof(ATOM_SIGNAL_FORMAT)-1, TRUE);
-    atomConnectorType   = MakeAtom(ATOM_CONNECTOR_TYPE,
-				   sizeof(ATOM_CONNECTOR_TYPE)-1, TRUE);
-    atomConnectorNumber = MakeAtom(ATOM_CONNECTOR_NUMBER,
-				   sizeof(ATOM_CONNECTOR_NUMBER)-1, TRUE);
-    atomOutputNumber    = MakeAtom(ATOM_OUTPUT_NUMBER,
-				   sizeof(ATOM_OUTPUT_NUMBER)-1, TRUE);
-
     randr = xnfcalloc(1, sizeof(struct rhdRandr));
     xf86CrtcConfigInit(pScrn, &rhdRRCrtcConfigFuncs);
     /* Maximum for D1GRPH_X/Y_END: 8k */
@@ -1016,6 +1129,7 @@ RHDRandrPreInit(ScrnInfoPtr pScrn)
     if (!xf86InitialConfiguration(pScrn, FALSE)) {
 	xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
 		   "RandR: No valid modes. Disabled.\n");
+	rhdPtr->randr = NULL;		/* TODO: not cleaning up correctly */
 	return FALSE;
     }
     xf86DrvMsg(pScrn->scrnIndex, X_INFO,
@@ -1026,12 +1140,30 @@ RHDRandrPreInit(ScrnInfoPtr pScrn)
     if (!xf86RandR12PreInit (pScrn)) {
 	xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
 		   "RandR: xf86RandR12PreInit failed. Disabled.\n");
-	rhdPtr->randr = NULL;
+	rhdPtr->randr = NULL;		/* TODO: not cleaning up correctly */
 	return FALSE;
     }
     return TRUE;
 }
 
+/* Wrapped PointerMoved for driver-level panning */
+static void
+rhdRRPointerMoved(int scrnIndex, int x, int y)
+{
+    ScrnInfoPtr pScrn  = xf86Screens[scrnIndex];
+    RHDPtr      rhdPtr = RHDPTR(xf86Screens[scrnIndex]);
+    int i;
+
+    for (i = 0; i < 2; i++) {
+	struct rhdCrtc *Crtc = rhdPtr->Crtc[i];
+	if (Crtc->scrnIndex == scrnIndex && Crtc->Active)
+	    rhdUpdateCrtcPos(Crtc, x + pScrn->frameX0, y + pScrn->frameY0);
+    }
+    UNWRAP_SCRNINFO(PointerMoved);
+    pScrn->PointerMoved(scrnIndex, x, y);
+    WRAP_SCRNINFO(PointerMoved, rhdRRPointerMoved); 
+}
+
 /* Call in ScreenInit after frame buffer + acceleration layers */
 Bool
 RHDRandrScreenInit(ScreenPtr pScreen)
@@ -1040,10 +1172,11 @@ RHDRandrScreenInit(ScreenPtr pScreen)
     RHDPtr rhdPtr = RHDPTR(pScrn);
 
     RHDFUNC(rhdPtr);
-    if (!xf86DiDGAInit(pScreen, (unsigned long) rhdPtr->FbBase))
-	return FALSE;
     if (!xf86CrtcScreenInit(pScreen))
 	return FALSE;
+    /* Wrap cursor for driver-level panning */
+    WRAP_SCRNINFO(PointerMoved, rhdRRPointerMoved); 
+
     RHDDebugRandrState(rhdPtr, "POST-ScreenInit");
     return TRUE;
 }
